;-*- Mode:LISP; Package:AP5; Base:10.; Syntax:Common-Lisp -*-
(in-package "AP5")
;
; commonlisp compatible approximations of try and fail
;
;; used to catch and throw multiple values

(defvar *failure* nil)
; We record the last failure because the commonlisp manual does not
; require that values of uncaught throws be accessible.

(declaim (special trystack))
(setq trystack nil)

(defmacro-displace fail (tag &rest values)
  (cond ((symbolp tag) (setq tag (kwote tag)))
	((and (listp tag)
	      (eq (car tag) 'quote)
	      (symbolp (cadr tag))))
	(t (error "bad throw tag ~A" tag)))
  `(progn (setq *failure* (list 'fail-marker ,@values))
	  (if (member ,tag trystack :test #'equal)
	      (throw ,tag *failure*)
	      (error "uncaught failure of type ~A:~&~A" ,tag (cdr *failure*)))))

#+ignore  ;; doesn't handle multiple-valued forms
(defmacro-displace try (form tag &rest code)
  (setq tag
	(cond ((symbolp tag) (kwote tag))
	      ((and (listp tag)
		    (eq (car tag) 'quote)
		    (symbolp (cadr tag)))
	       tag)			; allow either foo or 'foo why why why????
	      (t (Error "bad catch tag ~A" tag))))
  `(let* ((trystack (cons ,tag trystack))
	  (tryvalue
	   (catch ,tag ,form)))
     (cond ((eq (car (ilistp tryvalue)) 'fail-marker)
	    ;; you really want one protocol for direct THROWs and
            ;; something else for throws expressed through FAIL?
	    (setq tryvalue (cdr tryvalue))
	    ,@code)
	   (t tryvalue))))

(defmacro try (form tag &rest code)
  (setf tag
	(cond ((symbolp tag) (kwote tag))
	      ((and (listp tag)
		    (eq (car tag) 'quote)
		    (symbolp (second tag)))
	       tag)			; allow either foo or 'foo why why why????
	      (t (Error "bad catch tag ~A" tag))))
  `(let ((trystack (cons ,tag trystack)))
     (multiple-value-call
	 #'(lambda (&rest tryvalue)
	     (cond ((eq (car (ilistp (first tryvalue))) 'fail-marker)
		    ;; you really want one protocol for direct THROWs and
		    ;; something else for throws expressed through FAIL?
		    (setf tryvalue (rest (first tryvalue)))
		    ,@code)
		   (t (values-list tryvalue))))
       (catch ,tag ,form))))

; this seems related ...
(Defvar-resettable goals nil)
(defmacro with-goal (goal &rest body)
  `(let ((goals (cons ,goal goals))) ,.body))
